package de.lmu.ifi.mfchords.utils;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.*;

import android.app.Service;
import android.content.Intent;
import android.graphics.PointF;
import android.os.Environment;
import android.os.IBinder;
import android.support.v4.view.MotionEventCompat;
import android.util.Log;
import android.view.MotionEvent;
import net.sf.javaml.classification.KNearestNeighbors;
import net.sf.javaml.core.Dataset;
import net.sf.javaml.core.DefaultDataset;
import net.sf.javaml.core.DenseInstance;
import net.sf.javaml.core.Instance;



/**
 * The MFCRecognizer is the link between the classifier and the hand model
 * 
 */


public class MFCRecognizer extends Service{
	final static Charset ENCODING = Charset.defaultCharset();
	

	private ArrayList<Integer> pointer = new ArrayList<Integer>();
	//private static String PostureNone = "no posture recognized";
	
	private boolean classifierBuilt = false;
	private boolean feedForward = false;
	
	private ArrayList<RelativeMeasurement> rmMeasurements = new ArrayList<RelativeMeasurement>();
	private List<MFCListener> listeners = new ArrayList<MFCListener>();
	
	
	private HashMap<String, PointF> points = new HashMap<String, PointF>();
	
	private KNearestNeighbors knn = new KNearestNeighbors(3);
	private static Dataset normalizedData = new DefaultDataset();
	private HandModel handModel = new HandModel();
	
	
	private static MFCRecognizer mRecognizer = null;
	/**
	 * Singleton: access to the recognizer
	 * 
	 */
	public static MFCRecognizer getRecognizer(){
		return mRecognizer;
	}
	
	/**
	 * register your view to receive MFC events
	 * 
	 */
	public void addMFCRecognizer(MFCListener toAdd){
	   listeners.add(toAdd);
	}
	
	
	 
	private void MFCrecognized(String mfc) {
	   for (MFCListener hl : listeners) hl.MFCRecognized(mfc);
	
	}
	
	
	/**
	   * if set to 'true' it visualizes a feedforward of hovering fingers
	   * 
	   * @param wantFeedForward set this to yes if you want to visualize feedforward 
	   * 						about the approximate finger positions of the hovering fingers 
	   */
	public void wantFeedforward(boolean b){
		this.feedForward = b;
	}
	
	
	public void onCreate(){
		if(mRecognizer == null){
			mRecognizer = this;
		}
	}
	
	
	private void preProcess(ArrayList<PointF> pointers, boolean isTwoStepPosture){
		PointF basis_finger_one = null;
		PointF basis_finger_two = null;
		
		if(isTwoStepPosture){
			basis_finger_one = new PointF(pointers.get(0).x, pointers.get(0).y);
			basis_finger_two = new PointF(pointers.get(1).x, pointers.get(1).y);
			
			if(Math.abs(basis_finger_one.x - basis_finger_two.x) > 600 && Math.abs(basis_finger_one.y - basis_finger_two.y) < 400){
				//TP basis
				handModel.setChordBasis(Constants.ChordBasis_TP);
			}else{
				handModel.setChordBasis(Constants.ChordBasis_TI);
			}
			
		}else{
			handModel.setChordBasis(Constants.ChordBasis_NF);
		}
	}
	
	
	private void resetHandModel(){
		handModel.reset();
	}
	
	
	private class PointCompare implements Comparator<PointF> {
		
		PointCompare(){
			
		}
	
	   public int compare(PointF a, PointF b) {
		   if (a.x < b.x) {
			   return -1;
		   } else if (a.x > b.x) {
			   return 1;
		   } else {
			   return 0;
		   }
	   	}
}
	
	private ArrayList<PointF> orderTouches(ArrayList<PointF> points){
		List<PointF> list = new ArrayList<PointF>(points);
		/* adding points */
		Collections.sort(list,new PointCompare());
		
		return new ArrayList<PointF>(list);
	}
	
	
	private void processFeedforward(ArrayList<PointF> pointers){
		preProcess(pointers, true);
		ArrayList<PointF> orderedTouches = orderTouches(pointers);
		handModel.setBasis(orderedTouches);
		
		handModel.predictFingerPositions();
	}
	
	/**
	   * processPoints turns an array of touch points (PointF) into a RelativeMeasurement to feed the classifier or to 
	   * get it classified by the classifier.
	   * 
	   * @param pointers ArrayList of PointF points from the touch events.
	   * @param isTwoStepPosture The information of the view if the timing between the two touches
	   * 							of the basis and the third identifying touch is beyond threshold to 
	   * 							be interpreted as two-step posture  
	   * 
	   * @return RelativeMeasurement extracted relative measurements from the touch events.
	   */
	public RelativeMeasurement processPoints(ArrayList<PointF> pointers, boolean isTwoStepPosture){
		
		preProcess(pointers, isTwoStepPosture);
		
		//oder points
		ArrayList<PointF> orderedTouches = orderTouches(pointers);
		
		
		
		
		//extract body-relative features
		RelativeMeasurement relativeMeasures = new RelativeMeasurement();
		
		if(handModel.getChordBasis().equals(Constants.ChordBasis_TP) || handModel.getChordBasis().equals(Constants.ChordBasis_NF) ){
			relativeMeasures = handModel.GetRelativeMeasurements(orderedTouches.get(0), orderedTouches.get(2), orderedTouches.get(1));
		}else{ //if(handModel.chordBasis.equals(Constants.ChordBasis_TI)){
			relativeMeasures = handModel.GetRelativeMeasurements(orderedTouches.get(1), orderedTouches.get(0), orderedTouches.get(2));
		}
				
		return relativeMeasures;
		
	
		
		
	}
	
	
	private ArrayList<RelativeMeasurement> getRMForKey(String k){
		ArrayList<RelativeMeasurement> result = new ArrayList<RelativeMeasurement>();
		
		for(RelativeMeasurement rm : rmMeasurements){
			if(rm.getID().equals(k)){
				result.add(rm);
			}
		}
		
		return result;
	}
	
	private long twoFingersDownStamp = 0;
	public boolean onTouchEvent(MotionEvent event) {
		final int action = MotionEventCompat.getActionMasked(event);
		int index = MotionEventCompat.getActionIndex(event);
    	int pointerID = MotionEventCompat.getPointerId(event, index);
		
    	if(!pointer.contains(Integer.valueOf(pointerID))){
			pointer.add(Integer.valueOf(pointerID));
		}
    	
    	
    	
    	//PointF p = new PointF(event.getX(pointerID), event.getY(pointerID));
    	//points.put(pointerID, p);
    	
    	
    	
    	
    	
    	//points.append(pointerID, p);
    	//Log.v("myDebugTag", "points " + points.size());
		
		switch (action) {
    		case MotionEvent.ACTION_POINTER_DOWN:
    				if(pointer.size() == 2){
    					twoFingersDownStamp = System.currentTimeMillis();
    					
    					if(this.feedForward){
    						ArrayList<PointF> pointerList = new ArrayList<PointF>();
        					for(Integer i : pointer){
        						pointerList.add(new PointF(MotionEventCompat.getX(event, i),MotionEventCompat.getY(event, i)));
        					}
        					
        					processFeedforward(pointerList);
    					}
    					
    				}else if(pointer.size() == 3){
    					ArrayList<PointF> pointerList = new ArrayList<PointF>();
    					for(Integer i : pointer){
    						//Log.v("myDebugTag", "pointer " + MotionEventCompat.getX(event, i) + " " + MotionEventCompat.getY(event, i));
    						//feed the classifier
    						pointerList.add(new PointF(MotionEventCompat.getX(event, i),MotionEventCompat.getY(event, i)));
    					}
    					
    					boolean isTwoStepPosture = true;
    					if((System.currentTimeMillis() - twoFingersDownStamp) < 100){
    						isTwoStepPosture = false;
    					}
    					//rmMeasurements
    					RelativeMeasurement rm = processPoints(pointerList,isTwoStepPosture);
    					String predicted = getClassifiedMFC(rm);
    					
    					handModel.updateRemainingFingerPositions(predicted);
    					
    					/*for(Finger f : MFCRecognizer.getRecognizer().getHandModel().allFingers){
    						if(f!=null){
    							Log.v("myDebugTag", "finger " + f.name + " x " + f.position.x + " y " + f.position.y);
    							Log.v("myDebugTag", "predic " + f.name + " x " + f.getPredictedPosition().x + " y " + f.getPredictedPosition().y);
    						}
    						Log.v("myDebugTag", "-------");
    					}*/
    					
    					
    					Log.v("myDebugTag", "predicted " + predicted);
    					Log.v("myDebugTag", "\n \n calculated  " + rm.getEntry());
    					ArrayList<RelativeMeasurement> saveRM = getRMForKey(predicted);
    					for(RelativeMeasurement srm : saveRM){
    						Log.v("myDebugTag", "\n \n saved  " + srm.getEntry());
    					}
    					
    					
    					if(predicted != null){
    						for (MFCListener mfcl : listeners) {
                				mfcl.MFCRecognized(predicted);
                			}
    					}
    					
    				}else if(pointer.size() == 4){
    					Log.v("myDebugTag", "4 fingers ");
    					if(handModel.getRecognizedHandPosture() != Constants.Posture_None){
    						
    						PointF p = new PointF(event.getX(index), event.getY(index));
    						Log.v("myDebugTag", "checking  x" + p.x + " y " + p.y);
    						for(Finger f: handModel.getAllFingers()){
    							if(f.isTouching(p)){
    								//send out info
    								for (MFCListener mfcl : listeners) {
    	                				mfcl.MFCTapRecognized(handModel.getRecognizedHandPosture(),f.name);
    	                			}
    							}
    								
    						}
    						
    					}
    				}
    					
    				break;	
    				
    		case MotionEvent.ACTION_UP:
        		if(pointer.contains(Integer.valueOf(pointerID))){
        			pointer.remove(Integer.valueOf(pointerID));
        		}
        		
        		
        		for (MFCListener mfcl : listeners) {
    				mfcl.MFCReleased();
    			}
        		
        		resetHandModel();
        		break;
    		case MotionEvent.ACTION_POINTER_UP:
    			
        		if(pointer.contains(Integer.valueOf(pointerID))){
        			pointer.remove(Integer.valueOf(pointerID));
        		}
        		
        		PointF p = new PointF(event.getX(index), event.getY(index));
        		Log.v("myDebugTag", "checking  x" + p.x + " y " + p.y);
        		for(Finger f: handModel.getAllFingers()){
					if(f.isReleasing(p)){
						//send out info
						for (MFCListener mfcl : listeners) {
            				mfcl.MFCFingerReleased(f.name);
            			}
					}
						
				}
        		
        		break;
		}
		
		
		return true;
	}
	
	/**
	   * create classifier when the calibration process is done. 
	   *   
	   */
	public void createClassifier(){
		knn.buildClassifier(normalizedData);
		classifierBuilt = true;
	}
	
	/**
	   * feed the RelativeMeasurements to the Classifier if you want to calibrate. 
	   *   
	   */
	public void feedClassifier(RelativeMeasurement rm, String name){
		
		rm.setID(name);
		rmMeasurements.add(rm);
		
		
		double[] data = rm.getInstanceMeasures();
		Instance instance = new DenseInstance(data, name);
		normalizedData.add(instance);
		 
	}
	
	/**
	   * classifies the RelativeMeasurement using a KNN (N=3)
	   * 
	   * @param rm the relative measurement to classify
	   * @return a String with the name of the classified posture.  
	   */
	public String getClassifiedMFC(RelativeMeasurement rm){
		double[] d = rm.getInstanceMeasures();
		Instance i = new DenseInstance(d, "toClassify");
		String predicted = (String) knn.classify(i);
		handModel.setRecognizedHandPosture(predicted);
		return predicted;
	}
	
	
	
	
	public HandModel getHandModel() {
		return handModel;
	}

	/**
	   * builds classifier from calibration file (by default saved in external storage "defaultCalibration.txt").
	   * 
	   */
	public void readInCalibration() {
		
		File path = Environment.getExternalStoragePublicDirectory(
	            "NormalizedDataSets");
		File pathToFile = new File(path + "/"+ Constants.defaultFileName);
		
		
		FileInputStream is;
        Scanner scanner = null;
		try {
			is = new FileInputStream(pathToFile);
			scanner = new Scanner(is);
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
        
        if(scanner != null)
            Load(scanner);
    	
           knn = new KNearestNeighbors(3);
           knn.buildClassifier(normalizedData);
	}

	private void Load(Scanner scanner) {
		
		while (scanner.hasNextLine()){
			String line = scanner.nextLine();
			String[] data = line.split(";");
			
			double[] dInstance = new double[data.length-1];
			for(int i = 0; i< data.length-2 ;i++){
			    dInstance[i] = Double.valueOf(data[i]);
			    
			}
			
			
			Instance instance = new DenseInstance(dInstance,data[data.length-1]);
			Log.v("myDebugTag"," instance "+ instance.toString());
			normalizedData.add(instance);
		}
	    scanner.close();
	}
	
	@Override
	public IBinder onBind(Intent intent) {
		// TODO Auto-generated method stub
		return null;
	}
	
	@Override
	public int onStartCommand (Intent intent, int flags, int startId){
		
		//handleCommand(intent);
		return START_STICKY;
		
	}
	
	
	public boolean isClassifierBuilt(){
		return classifierBuilt;
	}
	
	
	/**
	   * writes calibration to file
	   * 
	   */
	public void writeCalibToFile(){
		
		boolean mExternalStorageAvailable = false;
		boolean mExternalStorageWriteable = false;
		String state = Environment.getExternalStorageState();

		if (Environment.MEDIA_MOUNTED.equals(state)) {
		    mExternalStorageAvailable = mExternalStorageWriteable = true;
		} else if (Environment.MEDIA_MOUNTED_READ_ONLY.equals(state)) {
		    mExternalStorageAvailable = true;
		    mExternalStorageWriteable = false;
		} else {
		    mExternalStorageAvailable = mExternalStorageWriteable = false;
		}
		
		if(!mExternalStorageWriteable || !mExternalStorageAvailable){
			Log.v("myDebugTag", "WARNING : cannot write to file");
		}
		
		File path = Environment.getExternalStoragePublicDirectory(
	            "NormalizedDataSets");
		
		
		
	    try {
	        path.mkdirs();
	        File file = new File(path, Constants.defaultFileName);
	        file.delete();
	       
	        FileWriter out = new FileWriter(file);
	        
	        
	        for(RelativeMeasurement rm: rmMeasurements){
	        	Log.v("myDebugTag", "RM " + rm.getEntry());
	        	out.write(rm.getEntry());
	        }
	        
	        
	        
            out.close();
	        
	        
	    } catch (IOException e) {
	        // Unable to create file, likely because external storage is
	        // not currently mounted.
	        Log.v("myDebugTag", "Error writing ", e);
	    }
		
	    
		
	}
	
}
